#include <assert.h>
#include <errno.h>
#include <math.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "pgc.h"
#include "pgc_lexer.h"
#include "pgc_parser.h"
#include "pgc_binding.h"
#include "pgc_evaluate.h"
#include "pgc_function.h"
#include "pgc_record.h"

extern int errno;

void pgc_evaluate_error (char *source, int error_code, int line_number, ...) {
  va_list argp;
  fprintf (stderr, "%s: ", source);
  switch (error_code) {
  case PGC_ERROR_LEXICAL:
    fprintf (stderr, "lexical error: ");
    break;
  case PGC_ERROR_UNEXPECTED_EOF:
    fprintf (stderr, "unexpected end of input: ");
    break;
  case PGC_ERROR_SYNTAX:
    fprintf (stderr, "syntax error: ");
    break;
  case PGC_ERROR_TYPE:
    fprintf (stderr, "type error: ");
    break;
  case PGC_ERROR_TYPE_NOT_COMPARABLE:
    va_start (argp, line_number);
    fprintf (stderr, "error; \'%s\' operator cannot compare this type: ",
	     va_arg (argp, char *));
    va_end (argp);
    break;
  case PGC_ERROR_DIVIDE_BY_ZERO:
    fprintf (stderr, "divide by zero: ");
    break;
  case PGC_ERROR_MODULO_ZERO:
    fprintf (stderr, "modulo zero: ");
    break;
  case PGC_ERROR_NAN:
    fprintf (stderr, "error; result is imaginary (NaN): ");
    break;
  case PGC_ERROR_VECTOR_TYPE:
    va_start (argp, line_number);
    fprintf (stderr, "wrong type for vector: "); 
    va_end (argp);
    break;
  case PGC_ERROR_SUBSCRIPT_RANGE:
    va_start (argp, line_number);
    fprintf (stderr, 
	     "subscript out of range for vector of length %d: ",
	     va_arg (argp, int));
    va_end (argp);
    break;
  case PGC_ERROR_SUBSCRIPT_TYPE:
    va_start (argp, line_number);
    fprintf (stderr, 
	     "subscript expression for vector must be of type int: ");
    va_end (argp);
    break;
  case PGC_ERROR_RECORD_ASSIGNMENT:
    fprintf (stderr, "error: record fields are read-only: ");
    break;
  case PGC_ERROR_RECORD_UNDEFINED_FIELD:
    va_start (argp, line_number);
    fprintf (stderr, "no field \'%s\' in definition for record type \'%s\': ",
	     va_arg (argp, char *), va_arg (argp, char *));
    va_end (argp);
    break;
  case PGC_ERROR_UNBOUND_IDENTIFIER:
    va_start (argp, line_number);
    fprintf (stderr, "unbound identifier \'%s\': ", va_arg (argp, char *));
    va_end (argp);
    break;
  case PGC_ERROR_FUNCTION_ARG_TYPE:
    va_start (argp, line_number);
    fprintf (stderr, "wrong argument type for function \'%s\': ",
	     va_arg (argp, char *));
    va_end (argp);
    break;
  case PGC_ERROR_FUNCTION_ARG_NUM:
    va_start (argp, line_number);
    fprintf (stderr, "wrong number of arguments for function \'%s\': ",
	     va_arg (argp, char *));
    va_end (argp);
    break;
  case PGC_ERROR_FUNCTION_UNDEFINED:
    va_start (argp, line_number);
    fprintf (stderr, "undefined function \'%s\': ", va_arg (argp, char *));
    va_end (argp);
    break;
  case PGC_ERROR_RESULT_ASSIGNMENT:
    fprintf (stderr, "cannot assign to result: ");
    break;
  default:
    fprintf (stderr, "unknown error: ");
    break;
  }
  fprintf (stderr, "line %d.\n", line_number);
  return;
}

/* TODO LIST:
   - "nil" and "result" should not be keywords -- they should be default
     entries in the binding table that cannot be redefined!  Or maybe result
     can be redefined. */

struct pgc_environment *evaluate 
(char *source, char *expression, struct pgc_environment *current_environment) {
  extern yyltype yylloc;
  struct yy_buffer_state *pgc_buffer_state = NULL;
  pgc_buffer_state = yy_expr_scan_string (expression);
  yylloc.first_line = 0;
  (void) yyparse ();
  yy_expr_delete_buffer (pgc_buffer_state);
  /* fprintf (stderr, "Parsing complete...\n"); */
  pgc_backup_result (current_environment);
  return typecheck (source, pgc_root, current_environment);
  
  /* Convert to list of lexical tokens... */
  /* Convert to parse tree */
  /* Type check */
  /* Add and return new bindings */
}

/* Clearly, if the type checker is going to add bindings, it has to perhaps
   return a value in the environment somehow -- perhaps this could be tacked
   onto the end -- then assignment statements could immediately claim it, other
   statements could just ignore it... hmm... but this seems to be a microcosm
   of the idea of storing the "result" in the env.  Let's say we put this
   interim val there. */

/* Other issues: fast lookups?  Should the bindings be arranged as a hash or
   not?  Is it difficult to resize a hash? */

struct pgc_environment *typecheck 
(char *source, struct pgc_parse_tree *parse_tree, 
 struct pgc_environment *current_environment) {
  /* We need to keep a backup of the current environment for instances in which
     we need to compare two child values, for example... */
  struct pgc_environment *new_environment = NULL;
  /* pgc_copy_environment (current_environment); */
  struct pgc_environment *temp_environment = NULL;
  struct pgc_environment *left_environment = NULL;
  struct pgc_environment *right_environment = NULL;
  int error_flag = 0;

  int scratch_type = -1;
  int i;
  void *scratch = NULL;
  void **scratch_args = NULL;
  
  switch (parse_tree->node_type) {
    /* This is the root type of a program... */
  case PGC_PARSE_EXPRESSION_SEQUENCE :
    switch (parse_tree->node_value_type) {
    case PGC_PARSE_EXPR:
       temp_environment = 
	 typecheck (source, parse_tree->node_value, current_environment);
       pgc_backup_result (temp_environment);
      break;
    case PGC_PARSE_VALUE_NIL:
      /* i.e., this is a nil expression... */
      temp_environment = pgc_copy_environment (current_environment);
      pgc_free_data (temp_environment->result_type, temp_environment->result);
      temp_environment->result_type = PGC_NIL;
      temp_environment->result = NULL;
    default:
      /* This is an error, probably... */
      break;
    }
    if (parse_tree->num_children == 1)
      temp_environment = 
	typecheck (source, parse_tree->children[0], temp_environment);
    new_environment = temp_environment;
    break;
  case PGC_PARSE_BINOP:
    switch (parse_tree->node_value_type) {
      /* These binops return a multitude of value types... */
    case PGC_PARSE_BINOP_PLUS:
      /* Let's overload PLUS so we can use it to concatenate strings. */
      left_environment = 
	typecheck (source, parse_tree->children[0], current_environment);
      right_environment = 
	typecheck (source, parse_tree->children[1],left_environment);
      if (left_environment->result_type == right_environment->result_type) {
	switch (left_environment->result_type) {
	case PGC_INT:
	  *((int *) right_environment->result) +=
	    *((int *) left_environment->result);
	  break;
	case PGC_FLOAT:
	  *((double *) right_environment->result) +=
	    *((double *) left_environment->result);
	  break;
	case PGC_STRING:
	  (char *) left_environment->result = 
	    realloc (left_environment->result, 
		     (strlen ((char *) left_environment->result) +
		      strlen ((char *) right_environment->result) + 1));
	  (char *) left_environment->result =
	    strcat ((char *) left_environment->result,
		    (char *) right_environment->result);
	  free (right_environment->result);
	  (char *) right_environment->result = 
	    strdup ((char *) left_environment->result);
	  break;
	default:
	  error_flag = 1;
	  break;
	}
      }
      else {
	if (left_environment->result_type == PGC_INT) {
	  if (right_environment->result_type == PGC_FLOAT)
	    *((double *) right_environment->result) +=
	      ((double) *((int *) left_environment->result));
	  else
	    error_flag = 1;
	}
	else {
	  if (left_environment->result_type == PGC_FLOAT) {
	    if (right_environment->result_type == PGC_INT) {
	      right_environment->result_type = PGC_FLOAT;
	      *((double *) left_environment->result) +=
		((double) *((int *) right_environment->result)); 
	      pgc_free_data (PGC_INT, right_environment->result);
	      right_environment->result = malloc (sizeof (double));
	      *((double *) right_environment->result) =
		*((double *) left_environment->result);
	    }
	    else
	      error_flag = 1;
	  }
	  else
	    error_flag = 1;
	}
      }
      pgc_delete_environment (left_environment);
      new_environment = right_environment;
      if (error_flag) {
	pgc_evaluate_error (source, PGC_ERROR_TYPE, parse_tree->line_number);
	pgc_free_data (new_environment->result_type, new_environment->result);
	new_environment->result_type = PGC_NIL;
	new_environment->result = NULL;
      }
      break;
    case PGC_PARSE_BINOP_MINUS:
      left_environment = 
	typecheck (source, parse_tree->children[0], current_environment);
      right_environment = 
	typecheck (source, parse_tree->children[1],left_environment);
      if (left_environment->result_type == right_environment->result_type) {
	switch (left_environment->result_type) {
	case PGC_INT:
	  *((int *) right_environment->result) =
	    *((int *) left_environment->result) - 
	    *((int *) right_environment->result);
	  break;
	case PGC_FLOAT:
	  *((double *) right_environment->result) =
	    *((double *) left_environment->result) -
	    *((double *) right_environment->result);
	  break;
	default:
	  error_flag = 1;
	  break;
	}
      }
      else {
	if (left_environment->result_type == PGC_INT) {
	  if (right_environment->result_type == PGC_FLOAT)
	    *((double *) right_environment->result) =
	      ((double) *((int *) left_environment->result)) -
	      *((double *) right_environment->result);
	  else
	    error_flag = 1;
	}
	else {
	  if (left_environment->result_type == PGC_FLOAT) {
	    if (right_environment->result_type == PGC_INT) {
	      right_environment->result_type = PGC_FLOAT;
	      *((double *) left_environment->result) -=
		((double) *((int *) right_environment->result)); 
	      pgc_free_data (PGC_INT, right_environment->result);
	      right_environment->result = malloc (sizeof (double));
	      *((double *) right_environment->result) =
		*((double *) left_environment->result);
	    }
	    else
	      error_flag = 1;
	  }
	  else
	    error_flag = 1;
	}
      }
      pgc_delete_environment (left_environment);
      new_environment = right_environment;
      if (error_flag) {
	pgc_evaluate_error (source, PGC_ERROR_TYPE, parse_tree->line_number);
	pgc_free_data (new_environment->result_type, new_environment->result);
	new_environment->result_type = PGC_NIL;
	new_environment->result = NULL;
      }
      break;
      /* leak-checked and non-new-env-reliant up to here... */
    case PGC_PARSE_BINOP_MULT:
      left_environment = 
	typecheck (source, parse_tree->children[0], current_environment);
      right_environment = 
	typecheck (source, parse_tree->children[1],left_environment);
      if (left_environment->result_type == right_environment->result_type) {
	switch (left_environment->result_type) {
	case PGC_INT:
	  *((int *) right_environment->result) *=
	    *((int *) left_environment->result);
	  break;
	case PGC_FLOAT:
	  if (*((double *) right_environment) == 0.0)
	    error_flag = 2;
	  *((double *) right_environment->result) *=
	    *((double *) left_environment->result);
	  break;
	default:
	  error_flag = 1;
	  break;
	}
      }
      else {
	if (left_environment->result_type == PGC_INT) {
	  if (right_environment->result_type == PGC_FLOAT)
	    *((double *) right_environment->result) *=
	      ((double) *((int *) left_environment->result));
	  else
	    error_flag = 1;
	}
	else {
	  if (left_environment->result_type == PGC_FLOAT) {
	    if (right_environment->result_type == PGC_INT) {
	      right_environment->result_type = PGC_FLOAT;
	      *((double *) left_environment->result) *=
		((double) *((int *) right_environment->result)); 
	      pgc_free_data (PGC_INT, right_environment->result);
	      right_environment->result = malloc (sizeof (double));
	      *((double *) right_environment->result) =
		*((double *) left_environment->result);
	    }
	    else
	      error_flag = 1;
	  }
	  else
	    error_flag = 1;
	}
      }
      pgc_delete_environment (left_environment);
      new_environment = right_environment;
      if (error_flag) {
	pgc_evaluate_error (source, PGC_ERROR_TYPE, parse_tree->line_number);
	pgc_free_data (new_environment->result_type, new_environment->result);
	new_environment->result_type = PGC_NIL;
	new_environment->result = NULL;
      }
      break;
    case PGC_PARSE_BINOP_DIV:
      left_environment = 
	typecheck (source, parse_tree->children[0], current_environment);
      right_environment = 
	typecheck (source, parse_tree->children[1],left_environment);
      if (left_environment->result_type == right_environment->result_type) {
	switch (left_environment->result_type) {
	case PGC_INT:
	  if (*((int *) right_environment->result) == 0) {
	    error_flag = 2;
	    break;
	  }
	  *((int *) right_environment->result) =
	    *((int *) left_environment->result) / 
	    *((int *) right_environment->result);
	  break;
	case PGC_FLOAT:
	  if (*((double *) right_environment->result) == 0.0) {
	    error_flag = 2;
	    break;
	  }
	  *((double *) right_environment->result) =
	    *((double *) left_environment->result) /
	    *((double *) right_environment->result);
	  break;
	default:
	  error_flag = 1;
	  break;
	}
      }
      else {
	if (left_environment->result_type == PGC_INT) {
	  if (right_environment->result_type == PGC_FLOAT) {
	    if (*((double *) right_environment->result) == 0.0)
	      error_flag = 2;
	    else {
	      *((double *) right_environment->result) =
		((double) *((int *) left_environment->result)) /
		*((double *) right_environment->result);
	    }
	  }
	  else
	    error_flag = 1;
	}
	else {
	  if (left_environment->result_type == PGC_FLOAT) {
	    if (right_environment->result_type == PGC_INT) {
	      if (*((int *) right_environment->result) == 0)
		error_flag = 2;
	      else {
		right_environment->result_type = PGC_FLOAT;
		*((double *) left_environment->result) /=
		  ((double) *((int *) right_environment->result)); 
		pgc_free_data (PGC_INT, right_environment->result);
		right_environment->result = malloc (sizeof (double));
		*((double *) right_environment->result) =
		  *((double *) left_environment->result);
	      }
	    }
	    else
	      error_flag = 1;
	  }
	  else
	    error_flag = 1;
	}
      }
      pgc_delete_environment (left_environment);
      new_environment = right_environment;
      switch (error_flag) {
      case 1:
	pgc_evaluate_error (source, PGC_ERROR_TYPE, parse_tree->line_number);
	break;
      case 2:
	pgc_evaluate_error (source, PGC_ERROR_DIVIDE_BY_ZERO, 
			    parse_tree->line_number);
	break;
      default:
	break;
      }
      if (error_flag) {
	pgc_free_data (new_environment->result_type, new_environment->result);
	new_environment->result_type = PGC_NIL;
	new_environment->result = NULL;
      }
      break;
    case PGC_PARSE_BINOP_MOD:
      left_environment = 
	typecheck (source, parse_tree->children[0], current_environment);
      right_environment = 
	typecheck (source, parse_tree->children[1], left_environment);
      if ((left_environment->result_type == PGC_INT) &&
	  (right_environment->result_type == PGC_INT)) {
	if (*((int *) right_environment->result) == 0)
	  error_flag = 2;
	else {
	  *((int *) right_environment->result) =
	    *((int *) left_environment->result) %
	    *((int *) right_environment->result);
	}
      }
      else 
	error_flag = 1;
      pgc_delete_environment (left_environment);
      new_environment = right_environment;
      switch (error_flag) {
      case 1:
	pgc_evaluate_error (source, PGC_ERROR_TYPE, parse_tree->line_number);
	break;
      case 2:
	pgc_evaluate_error (source, PGC_ERROR_MODULO_ZERO, 
			    parse_tree->line_number);
	break;
      default:
	break;
      }
      if (error_flag) {
	pgc_free_data (new_environment->result_type, new_environment->result);
	new_environment->result_type = PGC_NIL;
	new_environment->result = NULL;
      }      
      break;
    case PGC_PARSE_BINOP_POWER:
      left_environment = 
	typecheck (source, parse_tree->children[0],  current_environment);
      right_environment = 
	typecheck (source, parse_tree->children[1], left_environment);
      switch (left_environment->result_type) {
      case PGC_INT:
	switch (right_environment->result_type) {
	case PGC_INT:
	  temp_environment = pgc_copy_environment (right_environment);
	  pgc_free_data 
	    (temp_environment->result_type, temp_environment->result);
	  temp_environment->result_type = PGC_FLOAT;
	  temp_environment->result = malloc (sizeof (double));
	  *((double *) temp_environment->result) =
	    pow (*((int *) left_environment->result),
		 *((int *) right_environment->result));
	  pgc_delete_environment (right_environment);
	  right_environment = temp_environment;
	  break;
	case PGC_FLOAT:
	  *((double *) right_environment->result) =
	    pow (*((int *) left_environment->result),
		 *((double *) right_environment->result));
	  break;
	default:
	  error_flag = 1;
	  break;
	}
	break;
      case PGC_FLOAT:
	switch (right_environment->result_type) {
	case PGC_INT:
	  temp_environment = pgc_copy_environment (right_environment);
	  pgc_free_data 
	    (temp_environment->result_type, temp_environment->result);
	  temp_environment->result_type = PGC_FLOAT;
	  temp_environment->result = malloc (sizeof (double));
	  *((double *) temp_environment->result) =
	    pow (*((double *) left_environment->result),
		 *((int *) right_environment->result));
	  pgc_delete_environment (right_environment);
	  right_environment = temp_environment;
	  break;
	case PGC_FLOAT:
	  *((double *) right_environment->result) =
	    pow (*((double *) left_environment->result),
		 *((double *) right_environment->result));
	  break;
	default:
	  break;
	}
	break;
      default:
	error_flag = 1;
	break;
      }
      pgc_delete_environment (left_environment);
      new_environment = right_environment;
      if (error_flag) {
	pgc_evaluate_error (source, PGC_ERROR_TYPE, parse_tree->line_number);
	pgc_free_data (new_environment->result_type, new_environment->result);
	new_environment->result_type = PGC_NIL;
	new_environment->result = NULL;
      }
      else
	if (errno == EDOM) {
	  pgc_evaluate_error (source, PGC_ERROR_NAN, parse_tree->line_number);
	  pgc_free_data 
	    (new_environment->result_type, new_environment->result);
	  new_environment->result_type = PGC_NIL;
	  new_environment->result = NULL;
	}
      break;
  
      /* These binops return a boolean (int) value... */
      /* TODO: Allow mixing of INT with FLOAT... */
    case PGC_PARSE_BINOP_EQ:
      left_environment = 
	typecheck (source, parse_tree->children[0], current_environment);
      right_environment = 
	typecheck (source, parse_tree->children[1], left_environment);
      if (left_environment->result_type == right_environment->result_type) {
	switch (left_environment->result_type) {
	case PGC_INT:
	  if (*((int *) left_environment->result) ==
	      *((int *) right_environment->result))
	    *((int *) right_environment->result) = 1;
	  else
	    *((int *) right_environment->result) = 0;
	  break;
	case PGC_FLOAT:
	  temp_environment = pgc_copy_environment (right_environment);
	  pgc_free_data 
	    (temp_environment->result_type, temp_environment->result);
	  temp_environment->result_type = PGC_INT;
	  temp_environment->result = malloc (sizeof (int));
	  if (*((double *) left_environment->result) ==
	      *((double *) right_environment->result))
	    *((int *) temp_environment->result) = 1;
	  else
	    *((int *) temp_environment->result) = 0;
	  pgc_delete_environment (right_environment);
	  right_environment = temp_environment;
	  break;
	case PGC_STRING:
	  temp_environment = pgc_copy_environment (right_environment);
	  pgc_free_data 
	    (temp_environment->result_type, temp_environment->result);
	  temp_environment->result_type = PGC_INT;
	  temp_environment->result = malloc (sizeof (int));
	  if (strcmp ((char *) left_environment->result,
		      (char *) right_environment->result) == 0)
	    *((int *) temp_environment->result) = 1;
	  else
	    *((int *) temp_environment->result) = 0;
	  pgc_delete_environment (right_environment);
	  right_environment = temp_environment;
	  break;
	case PGC_NIL:
	  right_environment->result_type = PGC_INT;
	  right_environment->result = malloc (sizeof (int));
	  *((int *) right_environment->result) = 1;
	  break;
	default:
	  /* These are not comparable types! */
	  pgc_evaluate_error (source, PGC_ERROR_TYPE_NOT_COMPARABLE, 
			      parse_tree->line_number, "==");
	  error_flag = 1;
	  break;
	}
      }
      else {
	pgc_evaluate_error (source, PGC_ERROR_TYPE, parse_tree->line_number);
	error_flag = 1;
      }
      pgc_delete_environment (left_environment);
      new_environment = right_environment;
      if (error_flag) {
	pgc_free_data (new_environment->result_type, new_environment->result);
	new_environment->result_type = PGC_NIL;
	new_environment->result = NULL;
      }
      break;
    case PGC_PARSE_BINOP_NEQ:
      left_environment = 
	typecheck (source, parse_tree->children[0], current_environment);
      right_environment = 
	typecheck (source, parse_tree->children[1], left_environment);
      if (left_environment->result_type == right_environment->result_type) {
	switch (left_environment->result_type) {
	case PGC_INT:
	  if (*((int *) left_environment->result) !=
	      *((int *) right_environment->result))
	    *((int *) right_environment->result) = 1;
	  else
	    *((int *) right_environment->result) = 0;
	  break;
	case PGC_FLOAT:
	  temp_environment = pgc_copy_environment (right_environment);
	  pgc_free_data  
	    (temp_environment->result_type, temp_environment->result);
	  temp_environment->result_type = PGC_INT;
	  temp_environment->result = malloc (sizeof (int));
	  if (*((double *) left_environment->result) !=
	      *((double *) right_environment->result))
	    *((int *) temp_environment->result) = 1;
	  else
	    *((int *) temp_environment->result) = 0;
	  pgc_delete_environment (right_environment);
	  right_environment = temp_environment;
	  break;
	case PGC_STRING:
	  temp_environment = pgc_copy_environment (right_environment);
	  temp_environment->result_type = PGC_INT;
	  pgc_free_data 
	    (temp_environment->result_type, temp_environment->result);
	  temp_environment->result = malloc (sizeof (int));
	  if (strcmp ((char *) left_environment->result,
		      (char *) right_environment->result) != 0)
	    *((int *) temp_environment->result) = 1;
	  else
	    *((int *) temp_environment->result) = 0;
	  pgc_delete_environment (right_environment);
	  right_environment = temp_environment;
	  break;
	case PGC_NIL:
	  right_environment->result_type = PGC_INT;
	  right_environment->result = malloc (sizeof (int));
	  *((int *) right_environment->result) = 0;
	  break;
	default:
	  /* These are not comparable types! */
	  pgc_evaluate_error (source, PGC_ERROR_TYPE_NOT_COMPARABLE, 
			      parse_tree->line_number, "!=");
	  error_flag = 1;
	  break;
	}
      }
      else {
	pgc_evaluate_error (source, PGC_ERROR_TYPE, parse_tree->line_number);
	error_flag = 1;
      }
      pgc_delete_environment (left_environment);
      new_environment = right_environment;
      if (error_flag) {
	pgc_free_data (new_environment->result_type, new_environment->result);
	new_environment->result_type = PGC_NIL;
	new_environment->result = NULL;
      }
      break;
    case PGC_PARSE_BINOP_LT:
      left_environment = 
	typecheck (source, parse_tree->children[0], current_environment);
      right_environment = 
	typecheck (source, parse_tree->children[1], left_environment);
      if (left_environment->result_type == right_environment->result_type) {
	switch (left_environment->result_type) {
	case PGC_INT:
	  if (*((int *) left_environment->result) <
	      *((int *) right_environment->result))
	    *((int *) right_environment->result) = 1;
	  else
	    *((int *) right_environment->result) = 0;
	  break;
	case PGC_FLOAT:
	  temp_environment = pgc_copy_environment (right_environment);
	  pgc_free_data
	    (temp_environment->result_type, temp_environment->result);
	  temp_environment->result_type = PGC_INT;
	  temp_environment->result = malloc (sizeof (int));
	  if (*((double *) left_environment->result) <
	      *((double *) right_environment->result))
	    *((int *) temp_environment->result) = 1;
	  else
	    *((int *) temp_environment->result) = 0;
	  pgc_delete_environment (right_environment);
	  right_environment = temp_environment;
	  break;
	default:
	  /* These are not comparable types! */
	  pgc_evaluate_error (source, PGC_ERROR_TYPE_NOT_COMPARABLE, 
			      parse_tree->line_number, "<");
	  error_flag = 1;
	  break;
	}
      }
      else {
	pgc_evaluate_error (source, PGC_ERROR_TYPE, parse_tree->line_number);
	error_flag = 1;
      }
      pgc_delete_environment (left_environment);
      new_environment = right_environment;
      if (error_flag) {
	pgc_free_data (new_environment->result_type, new_environment->result);
	new_environment->result_type = PGC_NIL;
	new_environment->result = NULL;
      }
      break;
    case PGC_PARSE_BINOP_GT:
      left_environment = 
	typecheck (source, parse_tree->children[0], current_environment);
      right_environment = 
	typecheck (source, parse_tree->children[1], left_environment);
      if (left_environment->result_type == right_environment->result_type) {
	switch (left_environment->result_type) {
	case PGC_INT:
	  if (*((int *) left_environment->result) >
	      *((int *) right_environment->result))
	    *((int *) right_environment->result) = 1;
	  else
	    *((int *) right_environment->result) = 0;
	  break;
	case PGC_FLOAT:
	  temp_environment = pgc_copy_environment (right_environment);
	  pgc_free_data 
	    (temp_environment->result_type, temp_environment->result);
	  temp_environment->result_type = PGC_INT;
	  temp_environment->result = malloc (sizeof (int));
	  if (*((double *) left_environment->result) >
	      *((double *) right_environment->result))
	    *((int *) temp_environment->result) = 1;
	  else
	    *((int *) temp_environment->result) = 0;
	  pgc_delete_environment (right_environment);
	  right_environment = temp_environment;
	  break;
	default:
	  /* These are not comparable types! */
	  pgc_evaluate_error (source, PGC_ERROR_TYPE_NOT_COMPARABLE, 
			      parse_tree->line_number, ">");
	  error_flag = 1;
	  break;
	}
      }
      else {
	pgc_evaluate_error (source, PGC_ERROR_TYPE, parse_tree->line_number);
	error_flag = 1;
      }
      pgc_delete_environment (left_environment);
      new_environment = right_environment;
      if (error_flag) {
	pgc_free_data (new_environment->result_type, new_environment->result);
	new_environment->result_type = PGC_NIL;
	new_environment->result = NULL;
      }
      break;
    case PGC_PARSE_BINOP_LTE:
      left_environment = 
	typecheck (source, parse_tree->children[0], current_environment);
      right_environment = 
	typecheck (source, parse_tree->children[1], left_environment);
      if (left_environment->result_type == right_environment->result_type) {
	switch (left_environment->result_type) {
	case PGC_INT:
	  if (*((int *) left_environment->result) <=
	      *((int *) right_environment->result))
	    *((int *) right_environment->result) = 1;
	  else
	    *((int *) right_environment->result) = 0;
	  break;
	case PGC_FLOAT:
	  temp_environment = pgc_copy_environment (right_environment);
	  pgc_free_data 
	    (temp_environment->result_type, temp_environment->result);
	  temp_environment->result_type = PGC_INT;
	  temp_environment->result = malloc (sizeof (int));
	  if (*((double *) left_environment->result) <=
	      *((double *) right_environment->result))
	    *((int *) temp_environment->result) = 1;
	  else
	    *((int *) temp_environment->result) = 0;
	  pgc_delete_environment (right_environment);
	  right_environment = temp_environment;
	  break;
	default:
	  /* These are not comparable types! */
	  pgc_evaluate_error (source, PGC_ERROR_TYPE_NOT_COMPARABLE, 
			      parse_tree->line_number, "<=");
	  error_flag = 1;
	  break;
	}
      }
      else {
	pgc_evaluate_error (source, PGC_ERROR_TYPE, parse_tree->line_number);
	error_flag = 1;
      }
      pgc_delete_environment (left_environment);
      new_environment = right_environment;
      if (error_flag) {
	pgc_free_data (new_environment->result_type, new_environment->result);
	new_environment->result_type = PGC_NIL;
	new_environment->result = NULL;
      }
      break;
    case PGC_PARSE_BINOP_GTE:
      left_environment = 
	typecheck (source, parse_tree->children[0], current_environment);
      right_environment = 
	typecheck (source, parse_tree->children[1], left_environment);
      if (left_environment->result_type == right_environment->result_type) {
	switch (left_environment->result_type) {
	case PGC_INT:
	  if (*((int *) left_environment->result) >=
	      *((int *) right_environment->result))
	    *((int *) right_environment->result) = 1;
	  else
	    *((int *) right_environment->result) = 0;
	  break;
	case PGC_FLOAT:
	  temp_environment = pgc_copy_environment (right_environment);
	  pgc_free_data
	    (temp_environment->result_type, temp_environment->result);
	  temp_environment->result_type = PGC_INT;
	  temp_environment->result = malloc (sizeof (int));
	  if (*((double *) left_environment->result) >=
	      *((double *) right_environment->result))
	    *((int *) temp_environment->result) = 1;
	  else
	    *((int *) temp_environment->result) = 0;
	  pgc_delete_environment (right_environment);
	  right_environment = temp_environment;
	  break;
	default:
	  /* These are not comparable types! */
	  pgc_evaluate_error (source, PGC_ERROR_TYPE_NOT_COMPARABLE, 
			      parse_tree->line_number, ">=");
	  error_flag = 1;
	  break;
	}
      }
      else {
	pgc_evaluate_error (source, PGC_ERROR_TYPE, parse_tree->line_number);
	error_flag = 1;
      }
      pgc_delete_environment (left_environment);
      new_environment = right_environment;
      if (error_flag) {
	pgc_free_data (new_environment->result_type, new_environment->result);
	new_environment->result_type = PGC_NIL;
	new_environment->result = NULL;
      }
      break;
    default:
      /* Error -- it's not a handled type! */
      break;
    }
    break;

  case PGC_PARSE_FUNCTION:
    scratch = pgc_lookup_function (parse_tree->node_value);
    if (scratch != NULL) {
      char *function_name = parse_tree->node_value;
      parse_tree = parse_tree->children[0];
      new_environment = pgc_copy_environment (current_environment);
      if (parse_tree->node_value_type != PGC_PARSE_VALUE_NIL) {
	if (((struct pgc_function_definition *) scratch)->arg_count == 0) {
	  pgc_evaluate_error (source, PGC_ERROR_FUNCTION_ARG_NUM, 
			      parse_tree->line_number, function_name);
	  error_flag = 1;
	}
	else {
	  scratch_args = malloc 
	    (sizeof (void *) * 
	     ((struct pgc_function_definition *) scratch)->arg_count);
	  for (i = 0; 
	       i < (((struct pgc_function_definition *) scratch)->arg_count); 
	       i++) {
	    left_environment = 
	      typecheck (source, parse_tree->node_value, new_environment);
	    if (left_environment->result_type != 
		((struct pgc_function_definition *) scratch)->arg_types[i]) {
	      pgc_evaluate_error (source, PGC_ERROR_FUNCTION_ARG_TYPE, 
				  parse_tree->line_number,
				  function_name);
	      error_flag = 1;
	      { int j;
	        for (j = 0; j < i; j++)
		  /* The offending argument will not have been copied, so we
		     can use the types listed in arg_types to free () */
		  pgc_free_data 
		    (((struct pgc_function_definition *) scratch)->
		     arg_types[j], ((void **) scratch_args)[j]);
	      }
	      break;
	    }
	    scratch_args[i] = 
	      pgc_copy_data (((struct pgc_function_definition *) scratch)->
			     arg_types[i], 
			     left_environment->result);
	    if (i == ((struct pgc_function_definition *) scratch)->arg_count - 
		1) {
	      if (parse_tree->num_children != 0) {
		pgc_evaluate_error (source, PGC_ERROR_FUNCTION_ARG_NUM,
				    parse_tree->line_number, function_name);
		error_flag = 1;
		{ int j;
	          for (j = 0; j < i; j++) {
		    if (j == i)
		      pgc_free_data (left_environment->result_type, 
				     ((void **) scratch_args)[j]);
		    else
		      pgc_free_data 
			(((struct pgc_function_definition *) scratch)->
			 arg_types[j],
			 ((void **) scratch_args)[j]);
		  }
		}
		break;
	      }
	      else
		break;
	    }
	    else {
	      if (parse_tree->num_children == 0) {
		pgc_evaluate_error (source, PGC_ERROR_FUNCTION_ARG_NUM, 
				    parse_tree->line_number, function_name);
		error_flag = 1;
		{ int j;
	          for (j = 0; j < i; j++) {
		    if (j == i)
		      pgc_free_data (left_environment->result_type, 
				     ((void **) scratch_args)[j]);
		    else
		      pgc_free_data 
			(((struct pgc_function_definition *) scratch)->
			 arg_types[j],
			 ((void **) scratch_args)[j]);
		  }
		}
		break;
	      }
	      else
		parse_tree = parse_tree->children[0];
	    }
	    pgc_delete_environment (new_environment);
	    new_environment = left_environment;
	  }
	}
	if (!error_flag)
	  pgc_delete_environment (left_environment);
      }
      else {
	scratch_args = NULL;
	if (((struct pgc_function_definition *) scratch)->arg_count != 0) {
	  pgc_evaluate_error (source, PGC_ERROR_FUNCTION_ARG_NUM, 
			      parse_tree->line_number, function_name);
	  error_flag = 1;
	}
      }
      if (!error_flag) {
	pgc_free_data (new_environment->result_type, new_environment->result);
	new_environment->result_type = 
	  ((struct pgc_function_definition *) scratch)->return_type;
	/* scratch->function_pointer should point to a wrapper for the given
	   function... perhaps something we'll put in function_interfaces.h? */
	new_environment->result = 
	  (*(((struct pgc_function_definition *) scratch)->function_pointer))
	  (scratch_args);
	for (i = 0; 
	     i < (((struct pgc_function_definition *) scratch)->arg_count); 
	     i++)
	  pgc_free_data (((struct pgc_function_definition *) scratch)->
			 arg_types[i],
			 ((void **) scratch_args)[i]);
      }
    }
    else {
      pgc_evaluate_error (source, PGC_ERROR_FUNCTION_UNDEFINED, 
			  parse_tree->line_number, parse_tree->node_value);
      new_environment = pgc_copy_environment (current_environment);
      error_flag = 1;
    }
    if (error_flag) {
      pgc_free_data (new_environment->result_type, new_environment->result);
      new_environment->result_type = PGC_NIL;
      new_environment->result = NULL;
    }
    break;
  case PGC_PARSE_ASSIGNMENT:
    /* Important to distinguish between binding to identifiers and
       changing values within vectors. */
    right_environment = 
      typecheck (source, ((struct pgc_parse_tree *) parse_tree->children[1]), 
		 current_environment);
    if ((parse_tree->children[0]->node_value_type == 
	 PGC_PARSE_LVALUE_IDENTIFIER) &&
	(strcmp (parse_tree->children[0]->node_value, "result") == 0)) {
      pgc_evaluate_error (source, PGC_ERROR_RESULT_ASSIGNMENT, 
			  parse_tree->line_number);
      /* If the left side is an error, the right side shouldn't be
	 evaluated... */
      new_environment = pgc_copy_environment (current_environment);
    }
    else {
      if (parse_tree->children[0]->num_children == 0) {
	/* In this case, we're dealing with a variable binding... */
	scratch = malloc (sizeof (struct pgc_binding));
	((struct pgc_binding *) scratch)->binding_name = strdup
	  ((char *) ((struct pgc_parse_tree *) 
		     parse_tree->children[0])->node_value);
	((struct pgc_binding *) scratch)->value_type = 
	  right_environment->result_type;
	((struct pgc_binding *) scratch)->value = 
	  pgc_copy_data (right_environment->result_type, 
			 right_environment->result);
	((struct pgc_binding *) scratch)->reference_count = 1;
	new_environment = pgc_add_binding (scratch, right_environment);
	pgc_delete_environment (right_environment);
      }
      else {
	/* In this case, we're dealing with a vector_assignment... */
	parse_tree = parse_tree->children[0];
	scratch = 
	  pgc_lookup_binding (parse_tree->node_value, current_environment);
	if ((scratch != NULL) &&
	    ((((struct pgc_binding *) scratch)->value_type == PGC_RECORD) ||  
	     (((struct pgc_binding *) scratch)->value_type == PGC_VECTOR))) {
	  /* scratch_type = PGC_VECTOR; */
	  scratch = &(((struct pgc_binding *) scratch)->value);
	  new_environment = pgc_copy_environment (current_environment);
	  while (parse_tree->num_children == 1) {
	    parse_tree = parse_tree->children[0];	  
	    switch (parse_tree->node_value_type) {
	      /* After the top level of the lvalue, we shouldn't see any more
		 of PGC_PARSE_LVALUE_IDENTIFIER or whatever we're calling 
		 it. */
	    case PGC_PARSE_LVALUE_WITH_SUBSCRIPT:
	      left_environment = new_environment;
	      new_environment = 
		typecheck (source, parse_tree->node_value, left_environment);
	      pgc_delete_environment (left_environment);
	      if (scratch_type == PGC_VECTOR) {
		if (new_environment->result_type == PGC_INT) {
		  if ((*((int *) new_environment->result) < 
		       (*((struct pgc_vector **) scratch))->vector_length) &&
		      (*((int *) new_environment->result) >= 0)) {
		    scratch_type = 
		      (*((struct pgc_vector **) scratch))->vector_type;
		    scratch = 
		      &(((void **) (*((struct pgc_vector **) scratch))->
			 vector)[*((int *) new_environment->result)]);
		  }
		  else {
		    /* Error! */
		    pgc_evaluate_error 
		      (source, PGC_ERROR_SUBSCRIPT_RANGE, 
		       parse_tree->line_number,
		       (*((struct pgc_vector **) scratch))->vector_length);
		    error_flag = 1;
		    break;
		  }
		}
		else {
		  /* Error! */
		  pgc_evaluate_error (source, PGC_ERROR_SUBSCRIPT_TYPE, 
				      parse_tree->line_number);
		  error_flag = 1;
		  break;
		}
	      }
	      else {
		/* Error -- shouldn't try to subscript this! */
		pgc_evaluate_error (source, PGC_ERROR_TYPE, 
				    parse_tree->line_number);
		error_flag = 1;
		break;
	      }
	      break;
	    default:
	      /* And we're not allowing the user to assign to record fields --
		 they're read-only... */
	      pgc_evaluate_error (source, PGC_ERROR_RECORD_ASSIGNMENT, 
				  parse_tree->line_number);
	      error_flag = 1;
	      break;
	    }
	    if (error_flag)
	      break;
	  }
	  
	  /* Okay, we've reached the end of the line... time to assign the
	     value... */
	  if (scratch_type == right_environment->result_type)
	    *((void **) scratch) = 
	      pgc_copy_data (right_environment->result_type,
			     right_environment->result);
	  else 
	    /* Not the right value for that vector... */
	    pgc_evaluate_error (source, PGC_ERROR_TYPE, 
				parse_tree->line_number);
	}
	else
	  pgc_evaluate_error (source, PGC_ERROR_UNBOUND_IDENTIFIER, 
			      parse_tree->line_number,
			      parse_tree->node_value);
      }
    }
    pgc_free_data (new_environment->result_type, new_environment->result);
    new_environment->result_type = PGC_NIL;
    new_environment->result = NULL;
    break;
  case PGC_PARSE_LVALUE:
    /* Have to look up the binding... */
    scratch = pgc_lookup_binding 
      (parse_tree->node_value, current_environment);
    new_environment = pgc_copy_environment (current_environment);
    if (scratch != NULL) {
      if (parse_tree->num_children > 0) {
	if ((((struct pgc_binding *) scratch)->value_type == PGC_RECORD) || 
	    (((struct pgc_binding *) scratch)->value_type == PGC_VECTOR)) {
	  scratch_type = ((struct pgc_binding *) scratch)->value_type;
	  scratch = ((struct pgc_binding *) scratch)->value;
	  while (parse_tree->num_children == 1) {
	    parse_tree = parse_tree->children[0];	  
	    switch (parse_tree->node_value_type) {
	    case PGC_PARSE_LVALUE_IDENTIFIER:
	      /* After the top level of the lvalue, we shouldn't see any more
		 of PGC_PARSE_LVALUE_IDENTIFIER or whatever we're calling 
		 it. */
	      break;
	    case PGC_PARSE_LVALUE_WITH_SUBSCRIPT:
	      /* Here we're typechecking the subscript. */
	      left_environment = new_environment;
	      new_environment = 
		typecheck (source, parse_tree->node_value, left_environment);
	      pgc_delete_environment (left_environment);
	      if (scratch_type == PGC_VECTOR) {
		if (new_environment->result_type == PGC_INT) {
		  if ((*((int *) new_environment->result) < 
		       ((struct pgc_vector *) scratch)->vector_length) &&
		      (*((int *) new_environment->result) >= 0)) {
		    scratch_type = 
		      ((struct pgc_vector *) scratch)->vector_type;
		    scratch = ((void **) ((struct pgc_vector *) scratch)->
			       vector)[*((int *) new_environment->result)];
		  }
		  else {
		    /* Error! */
		    pgc_evaluate_error 
		      (source, PGC_ERROR_SUBSCRIPT_RANGE, 
		       parse_tree->line_number, 
		       ((struct pgc_vector *) scratch)->vector_length);
		    error_flag = 1;
		    break;
		  }
		}
		else {
		  /* Error! */
		  pgc_evaluate_error (source, PGC_ERROR_SUBSCRIPT_TYPE, 
				      parse_tree->line_number);
		  error_flag = 1;
		  break;
		}
	      }
	      else {
		/* Error -- shouldn't try to subscript this! */
		pgc_evaluate_error (source, PGC_ERROR_TYPE, 
				    parse_tree->line_number);
		error_flag = 1;
		break;
	      }
	      break;
	    case PGC_PARSE_LVALUE_WITH_FIELD:
	      if (scratch_type == PGC_RECORD) {
		struct pgc_record_definition *scratch_def =
		  pgc_lookup_record (((struct pgc_record *) scratch)->
				     record_type_name);
		if (scratch_def != NULL)
		  for (i = 0; i < scratch_def->field_count; i++) {
		    if (strcmp (scratch_def->public_field_names[i],
				parse_tree->node_value) == 0) {
		      scratch_type = scratch_def->public_field_types[i];
		      scratch = 
			((struct pgc_record *) scratch)->public_fields[i];
		      break;
		    }
		    else
		      if (i == scratch_def->field_count - 1) {
			pgc_evaluate_error 
			  (source, PGC_ERROR_RECORD_UNDEFINED_FIELD, 
			   parse_tree->line_number,
			   parse_tree->node_value,
			   ((struct pgc_record *) scratch)->record_type_name);
			error_flag = 1;
		      }
		  }
		else {
		  /* What happens if it does equal NULL? */
		  fprintf (stderr, 
			   "fatal: lvalue refers to undefined record type!\n");
		  assert (0);
		}
	      }
	      else {
		pgc_evaluate_error (source, PGC_ERROR_TYPE, 
				    parse_tree->line_number);
		error_flag = 1;
	      }
	    default:
	      break;
	    }
	  }
	  /* Okay, we've reached the value -- time to assign it to result... */
	  if (!error_flag) {
	    pgc_free_data 
	      (new_environment->result_type, new_environment->result);
	    new_environment->result_type = scratch_type;
	    new_environment->result = pgc_copy_data (scratch_type, scratch);
	  }
	}
	else {
	  /* Not the right value for that vector / field... */
	  pgc_evaluate_error (source, PGC_ERROR_TYPE, parse_tree->line_number);
	  error_flag = 1;
	}
      }
      else {
	/* If the parse tree has no children, then this must be a simple,
	   top-level request... */
	pgc_free_data (new_environment->result_type, new_environment->result);
	new_environment->result_type = 
	  ((struct pgc_binding *) scratch)->value_type;
	new_environment->result = pgc_copy_data 
	  (((struct pgc_binding *) scratch)->value_type,
	   ((struct pgc_binding *) scratch)->value);
      }
    }
    else {
      pgc_evaluate_error (source, PGC_ERROR_UNBOUND_IDENTIFIER, 
			  parse_tree->line_number,
			  parse_tree->node_value);
      error_flag = 1;
    }
    if (error_flag) {
      pgc_free_data (new_environment->result_type, new_environment->result);
      new_environment->result_type = PGC_NIL;
      new_environment->result = NULL;
    }
    break;

  case PGC_PARSE_IF_STATEMENT:
    left_environment = 
      typecheck (source, parse_tree->node_value, current_environment);
    if (left_environment->result_type == PGC_INT) {
      if (*((int *) (left_environment->result)) > 0)
	new_environment = 
	  typecheck (source, parse_tree->children[0], left_environment);
      else {
	new_environment = pgc_copy_environment (left_environment);
	pgc_free_data (new_environment->result_type, new_environment->result);
	new_environment->result_type = PGC_NIL;
	new_environment->result = NULL;	
      }
    }
    else {
      pgc_evaluate_error (source, PGC_ERROR_TYPE, parse_tree->line_number);
      new_environment = pgc_copy_environment (current_environment);
      pgc_free_data (new_environment->result_type, new_environment->result);
      new_environment->result_type = PGC_NIL;
      new_environment->result = NULL;
    }
    pgc_delete_environment (left_environment);
    break;
  case PGC_PARSE_IF_ELSE_STATEMENT:
    temp_environment = 
      typecheck (source, parse_tree->node_value, current_environment);
    if (temp_environment->result_type == PGC_INT) {
      left_environment = 
	typecheck (source, parse_tree->children[0], temp_environment);
      right_environment = 
	typecheck (source, parse_tree->children[1], temp_environment);
      if (left_environment->result_type != right_environment->result_type) {
	pgc_evaluate_error (source, PGC_ERROR_TYPE, parse_tree->line_number);
	new_environment = pgc_copy_environment (current_environment);
	pgc_free_data (new_environment->result_type, new_environment->result);
	new_environment->result_type = PGC_NIL;
	new_environment->result = NULL;
      }
      else {
	if (*((int *) (temp_environment->result)) != 0) {
	  new_environment = left_environment;
	  pgc_delete_environment (right_environment);
	}
	else {
	  new_environment = right_environment;
	  pgc_delete_environment (left_environment);
	}
      }
    }
    else {
      pgc_evaluate_error (source, PGC_ERROR_TYPE, 0);
      new_environment = pgc_copy_environment (current_environment);
      pgc_free_data (new_environment->result_type, new_environment->result);
      new_environment->result_type = PGC_NIL;
      new_environment->result = NULL;
    }
    pgc_delete_environment (temp_environment);
    break;
  case PGC_PARSE_FOREACH:
    /* This should be fun. */
    right_environment = 
      typecheck (source, parse_tree->children[0], current_environment);
    new_environment = pgc_copy_environment (right_environment);
    scratch = NULL;
    if (right_environment->result_type == PGC_VECTOR) {
      int i;
      pgc_free_data (new_environment->result_type, new_environment->result);
      new_environment->result_type = PGC_NIL;
      new_environment->result = NULL;
      for (i = 0; i < ((struct pgc_vector *) right_environment->result)->
	     vector_length; i++) {
	scratch = malloc (sizeof (struct pgc_binding));
	((struct pgc_binding *) scratch)->binding_name = 
	  strdup ((char *) parse_tree->node_value);
	((struct pgc_binding *) scratch)->value_type = 
	  ((struct pgc_vector *) right_environment->result)->vector_type;
	((struct pgc_binding *) scratch)->value = pgc_copy_data 
	  (((struct pgc_binding *) scratch)->value_type, 
	   ((void **) ((struct pgc_vector *) right_environment->result)->
	    vector)[i]);
	((struct pgc_binding *) scratch)->reference_count = 0;
	temp_environment = 
	  pgc_add_binding (((struct pgc_binding *) scratch), new_environment);
	pgc_delete_environment (new_environment);
	new_environment = 
	  typecheck (source, parse_tree->children[1], temp_environment);
	pgc_delete_environment (temp_environment);
      }
      pgc_delete_environment (right_environment);
    }
    else {
      pgc_evaluate_error (source, PGC_ERROR_TYPE, parse_tree->line_number);
      pgc_free_data (new_environment->result_type, new_environment->result);
      new_environment->result_type = PGC_NIL;
      new_environment->result = NULL;
    }
    break;
  case PGC_PARSE_DOWHILE:
    left_environment = current_environment;
    do {
      new_environment = 
	typecheck (source, parse_tree->children[0], left_environment);
      if (left_environment != current_environment)
	pgc_delete_environment (left_environment);
      right_environment = 
	typecheck (source, parse_tree->node_value, new_environment);
      pgc_delete_environment (new_environment);
      new_environment = right_environment;
      if (new_environment->result_type == PGC_INT) {
	if (*((int *) new_environment->result) == 0)
	  break;
	left_environment = new_environment;
      }
      else {
	pgc_evaluate_error (source, PGC_ERROR_TYPE, parse_tree->line_number);
	pgc_free_data (new_environment->result_type, new_environment->result);
	new_environment->result_type = PGC_NIL;
	new_environment->result = NULL;
	break;
      }
    }
    while (1);
    break;
  case PGC_PARSE_WHILEDO:
    left_environment = current_environment;
    while (1) {
      new_environment = 
	typecheck (source, parse_tree->node_value, left_environment);
      if (left_environment != current_environment)
	pgc_delete_environment (left_environment);
      if (new_environment->result_type == PGC_INT) {
	if (*((int *) new_environment->result) == 0)
	  break;
	else {
	  right_environment = 
	    typecheck (source, parse_tree->children[0], new_environment);
	  pgc_delete_environment (new_environment);
	  new_environment = right_environment;
	  left_environment = new_environment;
	}
      }
      else {
	pgc_evaluate_error (source, PGC_ERROR_TYPE, parse_tree->line_number);
	pgc_free_data (new_environment->result_type, new_environment->result);
	new_environment->result_type = PGC_NIL;
	new_environment->result = NULL;
	break;
      }
    }
    break;

  case PGC_PARSE_VALUE:
    new_environment = pgc_copy_environment (current_environment);
    pgc_free_data (new_environment->result_type, new_environment->result);
    switch (parse_tree->node_value_type) {
    case PGC_PARSE_VALUE_NIL:
      new_environment->result = NULL;
      new_environment->result_type = PGC_NIL;
      break;
    case PGC_PARSE_VALUE_INT:
      new_environment->result = 
	pgc_copy_data (PGC_INT, parse_tree->node_value);
      new_environment->result_type = PGC_INT;
      break;
    case PGC_PARSE_VALUE_FLOAT:
      new_environment->result =
	pgc_copy_data (PGC_FLOAT, parse_tree->node_value);
      new_environment->result_type = PGC_FLOAT;
      break;
    case PGC_PARSE_VALUE_STRING:
      new_environment->result = 
	pgc_copy_data (PGC_STRING, parse_tree->node_value);
      new_environment->result_type = PGC_STRING;
      break;
    case PGC_PARSE_VALUE_VECTOR:
      /* Okay, a little complicated -- have to make sure vector is all same
	 type... */	
      scratch = malloc (sizeof (struct pgc_vector));
      ((struct pgc_vector *) scratch)->vector_type = PGC_NIL;
      ((struct pgc_vector *) scratch)->vector_length = 0;
      ((struct pgc_vector *) scratch)->vector = NULL;
      right_environment = pgc_copy_environment (current_environment);
      if (parse_tree->num_children > 0) {
	int first_time_flag = 1;
	do {
	  parse_tree = parse_tree->children[0];
	  if (parse_tree->node_value_type == PGC_PARSE_VALUE_NIL)
	    break;
	  left_environment = 
	    typecheck (source, parse_tree->node_value, right_environment);
	  if (!first_time_flag) {
	    if (left_environment->result_type != 
		right_environment->result_type) {
	      pgc_evaluate_error (source, PGC_ERROR_VECTOR_TYPE, 
				  parse_tree->line_number);
	      error_flag = 1;
	      break;
	    }
	  }
	  first_time_flag = 0;
	  pgc_delete_environment (right_environment);
	  ((struct pgc_vector *) scratch)->vector_length++;
	  ((struct pgc_vector *) scratch)->vector = 
	    realloc (((struct pgc_vector *) scratch)->vector, 
		     (sizeof (void *) * 
		      ((struct pgc_vector *) scratch)->vector_length));
	  ((void **) ((struct pgc_vector *) scratch)->vector)
	    [((struct pgc_vector *) scratch)->vector_length - 1] =
	    pgc_copy_data (left_environment->result_type,
			   left_environment->result);
	  right_environment = left_environment;
	}
	while (parse_tree->num_children == 1);
	((struct pgc_vector *) scratch)->vector_type = 
	  right_environment->result_type;
      }
      new_environment = pgc_copy_environment (right_environment);
      /* We also delete left_environment when we do this... */
      pgc_delete_environment (right_environment);
      pgc_free_data (new_environment->result_type, new_environment->result);
      if (error_flag) {
	new_environment->result_type = PGC_NIL;
	new_environment->result = NULL;
      }
      else {
	new_environment->result_type = PGC_VECTOR;
	new_environment->result = scratch;
      }
      break;
    default:
      /* error! */
      break;
    }
    break;
  case PGC_PARSE_ERROR:
    /* As mentioned in pgc_parser.y, the node_value_type here comes from the 
       pgc.h enum pgc_error_types... */
    pgc_evaluate_error (source, parse_tree->node_value_type, 
			parse_tree->line_number);
    new_environment = pgc_copy_environment (current_environment);
    pgc_free_data (new_environment->result_type, new_environment->result);
    new_environment->result_type = PGC_NIL;
    new_environment->result = NULL;
  default:
    break;
  }
  return new_environment;
}



