/* Flex file for the pgc "language" */
%{
  #include <math.h>
  #include <string.h>
  #include "pgc_lexer.h"

  int commentLevel = 0;
  char *stringval = NULL;
  yyltype yylloc;
%}

DIGIT			[0-9]
IDENTIFIER		[\137a-zA-Z][\137a-zA-Z0-9]*

%option noyywrap prefix="yy_expr" outfile="pgc_lexer.c"
%s COMMENT STRING END_OF_THE_ROAD
%%

  /* C-style comments... */
<INITIAL>"/*"		{ commentLevel += 1;
                          BEGIN (COMMENT); }
  /* Newlines should not be part of this pattern -- they get matched later */
<COMMENT>(([^\/*\n])|(\/[^*\n])|("*"[^/\n]))+
<COMMENT>\n		{ yylloc.first_line += 1; }
<COMMENT>"*/"		{ commentLevel -= 1;
                          BEGIN (INITIAL); } 
<COMMENT><<EOF>> 	{ BEGIN (END_OF_THE_ROAD); 
			  return PGC_LEX_ERROR_EOF; }

  /* Strings... */
  /* The only type of string munging we do is allow for escaped quotes
     an newlines. */
<INITIAL>\"  				{ BEGIN (STRING);
					  stringval = malloc (sizeof (char));
					  stringval[0] = (char) NULL; } 
<STRING>([^\\\"\n])+			{ stringval = 
					    realloc (stringval, 
						     strlen (stringval) + 
						     strlen (yytext) + 1); 
					  stringval = 
					    strcat (stringval, yytext); }
<STRING>\\\"				{ stringval =
					    realloc (stringval,
						     strlen (stringval) + 1);
					  stringval = 
					    strcat (stringval, "\""); }
<STRING>\n				{ yylloc.first_line += 1; }
<STRING>\\n       			{ stringval = 
					    realloc (stringval, 
						     strlen (stringval) + 1);
					  stringval = 
					    strcat (stringval, "\n"); } 
<STRING>\\\\				{ stringval =
					    realloc (stringval,
						     strlen (stringval) + 1);
					  stringval = 
					    strcat (stringval, "\\"); }
<STRING>\"			{ yylval.stringval = stringval;
				  stringval = NULL;
               		          BEGIN (INITIAL);
				  return PGC_LEX_STRING; }
<STRING><<EOF>> 	{ BEGIN (END_OF_THE_ROAD);
			  return PGC_LEX_ERROR_EOF; }

<END_OF_THE_ROAD>.	{ BEGIN (INITIAL);
			  yyterminate (); }

  /* Put keywords before identifier... */
if			return PGC_LEX_IF;
else			return PGC_LEX_ELSE;
do			return PGC_LEX_DO;
while			return PGC_LEX_WHILE;
foreach			return PGC_LEX_FOREACH;
in			return PGC_LEX_IN;
nil			return PGC_LEX_NIL;

{DIGIT}+		{ yylval.intval = atoi (yytext); return PGC_LEX_INT; }
{DIGIT}*"."{DIGIT}+	{ yylval.floatval = atof (yytext); 
			  return PGC_LEX_FLOAT; }

  /* An identifier... */
{IDENTIFIER}		{ yylval.stringval = strdup (yytext); 
                          return PGC_LEX_IDENTIFIER; }

  /* The binary operations... */

"+"			return PGC_LEX_BINOP_PLUS;
"-"			return PGC_LEX_BINOP_MINUS;
"*"			return PGC_LEX_BINOP_MULT;
"/"			return PGC_LEX_BINOP_DIV;
"%"			return PGC_LEX_BINOP_MOD;
"^"			return PGC_LEX_BINOP_POWER;

"=="			return PGC_LEX_BINOP_EQ;
"!="			return PGC_LEX_BINOP_NEQ;
"<="			return PGC_LEX_BINOP_LTE;
">="			return PGC_LEX_BINOP_GTE;
"<"			return PGC_LEX_BINOP_LT;
">"			return PGC_LEX_BINOP_GT;

"="			return PGC_LEX_ASSIGN;
","			return PGC_LEX_COMMA;
";" 			return PGC_LEX_SEMICOLON;
"."			return PGC_LEX_DOT;

"("			return PGC_LEX_LPAREN;
")"			return PGC_LEX_RPAREN;
"["                     return PGC_LEX_LBRACK;
"]"			return PGC_LEX_RBRACK;

  /* Whitespace */

[\040\t]+
[\n]+			{ yylloc.first_line += (strlen (yytext)); }


<<EOF>> 		yyterminate ();
. 			{ BEGIN (END_OF_THE_ROAD); 
                          return PGC_LEX_ERROR; }

%%
